import struct

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class Satisfactory(ProtocolBase):
    """Satisfactory Protocol"""
    full_name = 'Satisfactory Protocol'

    async def get_status(self) -> dict:
        """
        Retrieves information about the server including state, version, and beacon port
        Server state: 1 - Idle (no game loaded), 2 - currently loading or creating a game, 3 - currently in game
        """
        # Credit: https://github.com/dopeghoti/SF-Tools/blob/main/Protocol.md

        # Send message id, protocol version
        request = struct.pack('2b', 0, 0) + 'opengsq'.encode()
        response = await SocketAsync.send_and_receive(self._host, self._port, self._timeout, request)
        br = BinaryReader(response)
        header = br.read_byte()

        if header != 1:
            raise InvalidPacketException('Packet header mismatch. Received: {}. Expected: {}.'.format(chr(header), chr(1)))

        br.read_byte()  # Protocol version
        br.read_bytes(8)  # Request data

        result = {}
        result['State'] = br.read_byte()
        result['Version'] = br.read_long()
        result['BeaconPort'] = br.read_short()

        return result


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        satisfactory = Satisfactory(host='delta3.ptse.host', port=15777, timeout=5.0)
        status = await satisfactory.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())



import re

from opengsq.binary_reader import BinaryReader
from opengsq.protocols.quake1 import Quake1


class Quake2(Quake1):
    """Quake2 Protocol"""
    full_name = 'Quake2 Protocol'

    def __init__(self, host: str, port: int, timeout: float = 5.0):
        """Quake2 Query Protocol"""
        super().__init__(host, port, timeout)
        self._response_header = 'print\n'

    def _parse_players(self, br: BinaryReader) -> list:
        players = []

        for matches in self._get_player_match_collections(br):
            matches: list[re.Match] = [match.group() for match in matches]

            player = {
                'frags': int(matches[0]),
                'ping': int(matches[1]),
            }

            if len(matches) > 2:
                player['name'] = str(matches[2]).strip('"')

            if len(matches) > 3:
                player['address'] = str(matches[3]).strip('"')

            players.append(player)

        return players


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        quake2 = Quake2(host='46.165.236.118', port=27910, timeout=5.0)
        status = await quake2.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())



from opengsq.protocols.source import Source


class WON(Source):
    """World Opponent Network (WON) Query Protocol"""
    full_name = 'World Opponent Network (WON) Protocol'

    _A2S_INFO = b'details\0'
    _A2S_PLAYER = b'players'
    _A2S_RULES = b'rules'


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        won = WON(host='212.227.190.150', port=27020, timeout=5.0)
        info = await won.get_info()
        players = await won.get_players()
        rules = await won.get_rules()
        print(json.dumps(info, indent=None, ensure_ascii=False) + '\n')
        print(json.dumps(players, indent=None, ensure_ascii=False) + '\n')
        print(json.dumps(rules, indent=None, ensure_ascii=False) + '\n')

    asyncio.run(main_async())



import struct


class BinaryReader:
    def __init__(self, data: bytes):
        self.__data = data
        self.stream_position = 0

    def length(self) -> int:
        return len(self.__data) - self.stream_position

    def is_end(self) -> int:
        return self.stream_position >= len(self.__data) - 1

    def prepend_bytes(self, data):
        self.__data = data + self.__data

    def read(self) -> bytes:
        return self.__data[self.stream_position:]

    def read_byte(self) -> int:
        data = self.__data[self.stream_position]
        self.stream_position += 1

        return data

    def read_bytes(self, count: int) -> bytes:
        data = self.__data[self.stream_position:self.stream_position + count]
        self.stream_position += count

        return data

    def read_short(self, unsigned=True) -> int:
        format = 'H' if unsigned else 'h'
        data = struct.unpack(f'<{format}', self.__data[self.stream_position:self.stream_position + 2])[0]
        self.stream_position += 2

        return data

    def read_long(self) -> int:
        data = struct.unpack('<l', self.__data[self.stream_position:self.stream_position + 4])[0]
        self.stream_position += 4

        return data

    def read_long_long(self) -> int:
        data = struct.unpack('<q', self.__data[self.stream_position:self.stream_position + 8])[0]
        self.stream_position += 8

        return data

    def read_float(self) -> float:
        data = struct.unpack('<f', self.__data[self.stream_position:self.stream_position + 4])[0]
        self.stream_position += 4

        return data

    def read_string(self, delimiters=[b'\x00'], encoding='utf-8', errors='ignore') -> str:
        bytes_string = b''

        while self.length() > 0:
            stream_byte = bytes([self.read_byte()])

            if stream_byte in delimiters:
                break

            bytes_string += stream_byte

        return str(bytes_string, encoding=encoding, errors=errors)



from opengsq.protocols import GameSpy3


class GameSpy4(GameSpy3):
    """GameSpy Protocol version 4"""
    full_name = 'GameSpy Protocol version 4'
    challenge = True


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        gs4 = GameSpy4(host='188.18.10.72', port=19133, timeout=5.0)
        server = await gs4.get_status()
        print(json.dumps(server, indent=None))

    asyncio.run(main_async())



import struct

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class Samp(ProtocolBase):
    """San Andreas Multiplayer Protocol"""
    full_name = 'San Andreas Multiplayer Protocol'

    _request_header = b'SAMP'
    _response_header = b'SAMP'

    async def get_status(self):
        br = await self.__send_and_receive(b'i')

        result = {}
        result['password'] = br.read_byte()
        result['numplayers'] = br.read_short()
        result['maxplayers'] = br.read_short()
        result['servername'] = self.__read_string(br, 4)
        result['gametype'] = self.__read_string(br, 4)
        result['language'] = self.__read_string(br, 4)

        return result

    async def get_players(self):
        """Server may not response when numplayers > 100"""
        br = await self.__send_and_receive(b'd')
        players = []
        numplayers = br.read_short()

        for _ in range(numplayers):
            player = {}
            player['id'] = br.read_byte()
            player['name'] = self.__read_string(br)
            player['score'] = br.read_long()
            player['ping'] = br.read_long()
            players.append(player)

        return players

    async def get_rules(self):
        br = await self.__send_and_receive(b'r')
        numrules = br.read_short()

        return dict((self.__read_string(br), self.__read_string(br)) for _ in range(numrules))

    async def __send_and_receive(self, data: bytes):
        # Format the address
        host = SocketAsync.gethostbyname(self._host)
        packet_header = struct.pack('BBBBH', *map(int, host.split('.') + [self._port])) + data
        request = self._request_header + packet_header

        # Validate the response
        response = await SocketAsync.send_and_receive(self._host, self._port, self._timeout, request)
        header = response[:len(self._response_header)]

        if header != self._response_header:
            raise InvalidPacketException(f'Packet header mismatch. Received: {header}. Expected: {self._response_header}.')

        return BinaryReader(response[len(self._response_header) + len(packet_header):])

    def __read_string(self, br: BinaryReader, read_offset=1):
        length = br.read_byte() if read_offset == 1 else br.read_long()
        return str(br.read_bytes(length), encoding='utf-8', errors='ignore')


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        samp = Samp(host='51.254.178.238', port=7777, timeout=5.0)
        status = await samp.get_status()
        print(json.dumps(status, indent=None) + '\n')
        players = await samp.get_players()
        print(json.dumps(players, indent=None) + '\n')
        rules = await samp.get_rules()
        print(json.dumps(rules, indent=None) + '\n')

    asyncio.run(main_async())



from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import ServerNotFoundException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync, SocketKind


class Scum(ProtocolBase):
    """Scum Protocol"""
    full_name = 'Scum Protocol'

    async def get_status(self, master_servers: list = None) -> dict:
        """
        Retrieves information about the server
        Notice: this method calls Scum.query_master_servers()
        function everytime if master_servers is not passed,
        you may need to cache the master servers if you had
        lots of servers to query.
        """
        ip = SocketAsync.gethostbyname(self._host)

        if master_servers is None:
            master_servers = await Scum.query_master_servers()

        for server in master_servers:
            if server['ip'] == ip and server['port'] == self._port:
                return server

        raise ServerNotFoundException()

    @staticmethod
    async def query_master_servers(host='172.107.16.215', port=1040) -> list:
        """
        Query SCUM Master-Server list
        """
        with SocketAsync(SocketKind.SOCK_STREAM) as sock:
            sock.settimeout(5)
            await sock.connect((host, port))
            sock.send(b'\x04\x03\x00\x00')

            total = -1
            response = b''
            servers = []

            while total == -1 or len(servers) < total:
                response += await sock.recv()
                br = BinaryReader(response)

                # first packet return the total number of servers
                if total == -1:
                    total = br.read_short()

                # server bytes length always 127
                while br.length() >= 127:
                    server = {}
                    server['ip'] = '.'.join(map(str, reversed([br.read_byte(), br.read_byte(), br.read_byte(), br.read_byte()])))
                    server['port'] = br.read_short()
                    server['name'] = str(br.read_bytes(100).rstrip(b'\x00'), encoding='utf-8', errors='ignore')
                    br.read_byte()  # skip
                    server['numplayers'] = br.read_byte()
                    server['maxplayers'] = br.read_byte()
                    server['time'] = br.read_byte()
                    br.read_byte()  # skip
                    server['password'] = ((br.read_byte() >> 1) & 1) == 1
                    br.read_bytes(7)  # skip
                    v = list(reversed([hex(br.read_byte())[2:].rjust(2, '0') for _ in range(8)]))
                    server['version'] = f'{int(v[0], 16)}.{int(v[1], 16)}.{int(v[2] + v[3], 16)}.{int(v[4] + v[5] + v[6] + v[7], 16)}'
                    servers.append(server)

                # if the length is less than 127, save the unused bytes for next loop
                response = br.read()

        return servers


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        scum = Scum(host='15.235.181.19', port=7042, timeout=5.0)
        master_servers = await scum.query_master_servers()
        server = await scum.get_status(master_servers)
        print(json.dumps(server, indent=None) + '\n')

    asyncio.run(main_async())



import bz2
import random
import zlib
from enum import Enum

from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import AuthenticationException, InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync, SocketKind


class Source(ProtocolBase):
    """Source Engine Protocol"""
    full_name = 'Source Engine Protocol'

    _A2S_INFO = b'\x54Source Engine Query\0'
    _A2S_PLAYER = b'\x55'
    _A2S_RULES = b'\x56'

    class __RequestHeader():
        A2A_PING = b'\x69'
        A2S_SERVERQUERY_GETCHALLENGE = b'\x57'

    class __ResponseHeader():
        S2C_CHALLENGE = 0x41
        S2A_INFO_SRC = 0x49
        S2A_INFO_DETAILED = 0x6D
        S2A_PLAYER = 0x44
        S2A_RULES = 0x45
        A2A_ACK = 0x6A

    async def get_info(self) -> dict:
        """
        Retrieves information about the server including,
        but not limited to: its name, the map currently being played, and the number of players.

        See: https://developer.valvesoftware.com/wiki/Server_queries#A2S_INFO
        """
        response_data = await self.__connect_and_send_challenge(self._A2S_INFO)

        br = BinaryReader(response_data)
        header = br.read_byte()

        if header != self.__ResponseHeader.S2A_INFO_SRC and header != self.__ResponseHeader.S2A_INFO_DETAILED:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {} or {}.'
                .format(chr(header), chr(self.__ResponseHeader.S2A_INFO_SRC), chr(self.__ResponseHeader.S2A_INFO_DETAILED))
            )

        if header == self.__ResponseHeader.S2A_INFO_SRC:
            return self.__parse_from_info_src(br)

        # Obsolete GoldSource Response
        return self.__parse_from_info_detailed(br)

    def __parse_from_info_src(self, br: BinaryReader) -> dict:
        info = {}
        info['Protocol'] = br.read_byte()
        info['Name'] = br.read_string()
        info['Map'] = br.read_string()
        info['Folder'] = br.read_string()
        info['Game'] = br.read_string()
        info['ID'] = br.read_short()
        info['Players'] = br.read_byte()
        info['MaxPlayers'] = br.read_byte()
        info['Bots'] = br.read_byte()
        info['ServerType'] = chr(br.read_byte())
        info['Environment'] = chr(br.read_byte())
        info['Visibility'] = br.read_byte()
        info['VAC'] = br.read_byte()

        # These fields only exist in a response if the server is running The Ship
        if info['ID'] == 2400:
            info['Mode'] = br.read_byte()
            info['Witnesses'] = br.read_byte()
            info['Duration'] = br.read_byte()

        info['Version'] = br.read_string()
        info['EDF'] = br.read_byte()

        if info['EDF'] & 0x80:
            info['GamePort'] = br.read_short()

        if info['EDF'] & 0x10:
            info['SteamID'] = br.read_long_long()

        if info['EDF'] & 0x40:
            info['SpecPort'] = br.read_short()
            info['SpecName'] = br.read_string()

        if info['EDF'] & 0x20:
            info['Keywords'] = br.read_string()

        if info['EDF'] & 0x01:
            info['GameID'] = br.read_long_long()

        return info

    def __parse_from_info_detailed(self, br: BinaryReader) -> dict:
        info = {}
        info['Address'] = br.read_string()
        info['Name'] = br.read_string()
        info['Map'] = br.read_string()
        info['Folder'] = br.read_string()
        info['Game'] = br.read_string()
        info['Players'] = br.read_byte()
        info['MaxPlayers'] = br.read_byte()
        info['Protocol'] = br.read_byte()
        info['ServerType'] = chr(br.read_byte())
        info['Environment'] = chr(br.read_byte())
        info['Visibility'] = br.read_byte()
        info['Mod'] = br.read_byte()

        if info['Mod'] == 1:
            info['Link'] = br.read_string()
            info['DownloadLink'] = br.read_string()

            br.read_byte()

            info['Version'] = br.read_long()
            info['Size'] = br.read_long()
            info['Type'] = br.read_byte()
            info['DLL'] = br.read_byte()

        info['VAC'] = br.read_byte()
        info['Bots'] = br.read_byte()

        return info

    async def get_players(self) -> list:
        """
        This query retrieves information about the players currently on the server.

        See: https://developer.valvesoftware.com/wiki/Server_queries#A2S_PLAYER
        """
        response_data = await self.__connect_and_send_challenge(self._A2S_PLAYER)

        br = BinaryReader(response_data)
        header = br.read_byte()

        if header != self.__ResponseHeader.S2A_PLAYER:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(self.__ResponseHeader.S2A_PLAYER))
            )

        player_count = br.read_byte()
        players = []

        for _ in range(player_count):
            br.read_byte()

            player = {}
            player['Name'] = br.read_string()
            player['Score'] = br.read_long()
            player['Duration'] = br.read_float()
            players.append(player)

        if br.length() > 0:
            for i in range(player_count):
                players[i]['Deaths'] = br.read_long()
                players[i]['Money'] = br.read_long()

        return players

    async def get_rules(self) -> dict:
        """
        Returns the server rules, or configuration variables in name/value pairs.

        See: https://developer.valvesoftware.com/wiki/Server_queries#A2S_RULES
        """
        response_data = await self.__connect_and_send_challenge(self._A2S_RULES)

        br = BinaryReader(response_data)
        header = br.read_byte()

        if header != self.__ResponseHeader.S2A_RULES:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(chr(header), chr(self.__ResponseHeader.S2A_RULES))
            )

        rule_count = br.read_short()
        rules = dict((br.read_string(), br.read_string()) for _ in range(rule_count))

        return rules

    async def __connect_and_send_challenge(self, header: __RequestHeader) -> bytes:
        # Connect to remote host
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # Send and receive
            request_base = b'\xFF\xFF\xFF\xFF' + header
            request_data = request_base

            if header != self._A2S_INFO:
                request_data += b'\xFF\xFF\xFF\xFF'

            sock.send(request_data)

            # Retries 3 times, some servers require multiple challenges
            for _ in range(3):
                response_data = await self.__receive(sock)
                br = BinaryReader(response_data)
                header = br.read_byte()

                # The server response with a challenge
                if header == self.__ResponseHeader.S2C_CHALLENGE:
                    challenge = br.read()

                    # Send the challenge and receive
                    sock.send(request_base + challenge)
                else:
                    break

        return response_data

    async def __receive(self, sock: SocketAsync) -> bytes:
        total_packets = -1
        payloads = dict()
        packets = list()

        while True:
            response_data = await sock.recv()
            packets.append(response_data)

            br = BinaryReader(response_data)
            header = br.read_long()

            # Simple Response Format
            if header == -1:
                # Return the payload
                return br.read()

            # Packet id
            id = br.read_long()
            is_compressed = (id & 0x80000000) < 0

            # Check is GoldSource multi-packet response format
            if self.__is_gold_source_split(BinaryReader(br.read())):
                return await self.__parse_gold_source_packet(sock, packets)

            # The total number of packets
            total_packets = br.read_byte()

            # The number of the packet
            number = br.read_byte()

            # Packet size
            br.read_short()

            if number == 0 and is_compressed:
                # Decompressed size
                br.read_long()

                # CRC32 sum
                crc32_checksum = br.read_long()

            payloads[number] = br.read()

            if total_packets == -1 or len(payloads) < total_packets:
                continue

            break

        # Combine the payloads
        combined_payload = b''.join(payloads[number] for number in sorted(payloads))

        # Decompress the payload
        if is_compressed:
            combined_payload = bz2.decompress(combined_payload)

            # Check CRC32 sum
            if zlib.crc32(combined_payload) != crc32_checksum:
                raise InvalidPacketException('CRC32 checksum mismatch of uncompressed packet data.')

        return combined_payload.startswith(b'\xFF\xFF\xFF\xFF') and combined_payload[4:] or combined_payload

    def __is_gold_source_split(self, br: BinaryReader):
        # Upper 4 bits represent the number of the current packet (starting at 0)
        number = br.read_byte() >> 4

        # Check is it Gold Source packet split format
        return number == 0 and br.read().startswith(b'\xFF\xFF\xFF\xFF')

    async def __parse_gold_source_packet(self, sock: SocketAsync, packets: list):
        total_packets = -1
        payloads = dict()

        while total_packets == -1 or len(payloads) < total_packets:
            # Load the old received packets first, then receive the packets
            if len(payloads) < len(packets):
                response_data = packets[len(payloads)]
            else:
                response_data = await sock.recv()

            br = BinaryReader(response_data)

            # Header
            br.read_long()

            # Packet id
            br.read_long()

            # The total number of packets
            total_packets = br.read_byte()

            # Upper 4 bits represent the number of the current packet (starting at 0)
            number = total_packets >> 4

            # Bottom 4 bits represent the total number of packets (2 to 15)
            total_packets &= 0x0F

            payloads[number] = br.read()

        # Combine the payloads
        combined_payload = b''.join(payloads[number] for number in sorted(payloads))

        return combined_payload.startswith(b'\xFF\xFF\xFF\xFF') and combined_payload[4:] or combined_payload

    class RemoteConsole(ProtocolBase):
        """Source RCON Protocol"""

        full_name = 'Source RCON Protocol'

        def __init__(self, host: str, port: int = 27015, timeout: float = 5.0):
            """Source RCON Protocol"""
            super().__init__(host, port, timeout)

            self._sock = None

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            self.close()

        def close(self):
            """Close the connection"""
            if self._sock:
                self._sock.close()

        async def authenticate(self, password: str):
            """Authenticate the connection"""

            # Connect
            self._sock = SocketAsync(SocketKind.SOCK_STREAM)
            self._sock.settimeout(self._timeout)
            await self._sock.connect((self._host, self._port))

            # Send password
            id = random.randrange(4096)
            self._sock.send(self.__Packet(id, self.__PacketType.SERVERDATA_AUTH.value, password).get_bytes())

            # Receive and parse as Packet
            response_data = await self._sock.recv()
            packet = self.__Packet(response_data)

            # Sometimes it will return a PacketType.SERVERDATA_RESPONSE_VALUE, so receive again
            if packet.type != self.__PacketType.SERVERDATA_AUTH_RESPONSE.value:
                response_data = await self._sock.recv()
                packet = self.__Packet(response_data)

            # Throw exception if not PacketType.SERVERDATA_AUTH_RESPONSE
            if packet.type != self.__PacketType.SERVERDATA_AUTH_RESPONSE.value:
                self._sock.close()
                raise InvalidPacketException(
                    'Packet header mismatch. Received: {}. Expected: {}.'
                    .format(chr(packet.type), chr(self.__PacketType.SERVERDATA_AUTH_RESPONSE.value))
                )

            # Throw exception if authentication failed
            if packet.id == -1 or packet.id != id:
                self._sock.close()
                raise AuthenticationException('Authentication failed')

        async def send_command(self, command: str):
            """Send command to the server"""

            # Send the command and a empty command packet
            id = random.randrange(4096)
            dummy_id = id + 1
            self._sock.send(self.__Packet(id, self.__PacketType.SERVERDATA_EXECCOMMAND.value, command).get_bytes())
            self._sock.send(self.__Packet(dummy_id, self.__PacketType.SERVERDATA_EXECCOMMAND.value, '').get_bytes())

            packet_bytes = bytes([])
            response = ''

            while True:
                # Receive
                response_data = await self._sock.recv()

                # Concat to last unused bytes
                packet_bytes += response_data

                # Get the packets and get the unused bytes
                packets, packet_bytes = self.__get_packets(packet_bytes)

                # Loop all packets
                for packet in packets:
                    if packet.id == dummy_id:
                        return response

                    response += packet.body

        # Handle Multiple-packet Responses
        def __get_packets(self, packet_bytes: bytes):
            packets = []

            br = BinaryReader(packet_bytes)

            # + 4 to ensure br.ReadInt32() is readable
            while br.stream_position + 4 < len(packet_bytes):
                size = br.read_long()

                if br.stream_position + size > len(packet_bytes):
                    return packets, packet_bytes[br.stream_position - 4:]

                id = br.read_long()
                type = br.read_long()
                body = br.read_string()
                br.read_byte()

                packets.append(self.__Packet(id, type, body))

            return packets, bytes([])

        class __PacketType(Enum):
            SERVERDATA_AUTH = 3
            SERVERDATA_AUTH_RESPONSE = 2
            SERVERDATA_EXECCOMMAND = 2
            SERVERDATA_RESPONSE_VALUE = 0

        class __Packet:
            def __init__(self, *args):
                if len(args) == 3:
                    self.id = args[0]
                    self.type = args[1]
                    self.body = args[2]
                else:
                    # Single-packet Responses
                    br = BinaryReader(args[0])
                    br.read_long()
                    self.id = br.read_long()
                    self.type = br.read_long()
                    self.body = br.read_string()

            def get_bytes(self):
                packet_bytes = self.id.to_bytes(4, byteorder='little')
                packet_bytes += self.type.to_bytes(4, byteorder='little')
                packet_bytes += str.encode(self.body + '\0')
                return len(packet_bytes).to_bytes(4, byteorder='little') + packet_bytes


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        source = Source(host='209.205.114.187', port=27015, timeout=5.0)
        info = await source.get_info()
        print(json.dumps(info, indent=None, ensure_ascii=False) + '\n')
        players = await source.get_players()
        print(json.dumps(players, indent=None, ensure_ascii=False) + '\n')
        rules = await source.get_rules()
        print(json.dumps(rules, indent=None, ensure_ascii=False) + '\n')

    asyncio.run(main_async())



from enum import Flag, auto

from opengsq.binary_reader import BinaryReader
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class GameSpy2(ProtocolBase):
    """GameSpy Protocol version 2"""
    full_name = 'GameSpy Protocol version 2'

    class Request(Flag):
        INFO = auto()
        PLAYERS = auto()
        TEAMS = auto()

    async def get_status(self, request: Request = Request.INFO | Request.PLAYERS | Request.TEAMS) -> dict:
        """Retrieves information about the server including, Info, Players, and Teams."""
        # Connect to remote host
        with SocketAsync() as sock:
            sock.settimeout(self._timeout)
            await sock.connect((self._host, self._port))

            # Send Request
            sock.send(b'\xFE\xFD\x00\x04\x05\x06\x07' + self.__get_request_bytes(request))

            # Server response
            response = await sock.recv()

        # Remove the first 5 bytes { 0x00, 0x04, 0x05, 0x06, 0x07 }
        br = BinaryReader(response[5:])

        status = {}

        if self.__has_flag(request, self.Request.INFO):
            status['info'] = self.__get_info(br)

        if self.__has_flag(request, self.Request.PLAYERS):
            status['players'] = self.__get_players(br)

        if self.__has_flag(request, self.Request.TEAMS):
            status['teams'] = self.__get_teams(br)

        return status

    def __get_request_bytes(self, request: Request):
        request_bytes = self.__has_flag(request, self.Request.INFO) and b'\xFF' or b'\x00'
        request_bytes += self.__has_flag(request, self.Request.PLAYERS) and b'\xFF' or b'\x00'
        request_bytes += self.__has_flag(request, self.Request.TEAMS) and b'\xFF' or b'\x00'

        return request_bytes

    def __has_flag(self, request, flag) -> bool:
        return request & flag == flag

    def __get_info(self, br: BinaryReader) -> dict:
        info = {}

        # Read all key values
        while br.length() > 0:
            key = br.read_string()

            if key == '':
                break

            info[key] = br.read_string().strip()

        return info

    def __get_players(self, br: BinaryReader) -> list:
        players = []

        # Skip a byte
        br.read_byte()

        # Get player count
        player_count = br.read_byte()

        # Get all keys
        keys = []

        while br.length() > 0:
            key = br.read_string()

            if key == '':
                break

            keys.append(key.rstrip('_'))

        # Set all keys and values
        for _ in range(player_count):
            players.append({key: br.read_string().strip() for key in keys})

        return players

    def __get_teams(self, br: BinaryReader) -> list:
        teams = []

        # Skip a byte
        br.read_byte()

        # Get team count
        team_count = br.read_byte()

        # Get all keys
        keys = []

        while br.length() > 0:
            key = br.read_string()

            if key == '':
                break

            keys.append(key.rstrip('t').rstrip('_'))

        # Set all keys and values
        for _ in range(team_count):
            teams.append({key: br.read_string().strip() for key in keys})

        return teams


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        gs2 = GameSpy2(host='158.69.118.94', port=23000, timeout=5.0)
        status = await gs2.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())



from opengsq.binary_reader import BinaryReader
from opengsq.exceptions import InvalidPacketException
from opengsq.protocol_base import ProtocolBase
from opengsq.socket_async import SocketAsync


class Raknet(ProtocolBase):
    """Raknet Protocol (https://wiki.vg/Raknet_Protocol)"""
    full_name = 'Raknet Protocol'

    __ID_UNCONNECTED_PING = b'\x01'
    __ID_UNCONNECTED_PONG = b'\x1C'
    __TIMESTAMP = b'\x12\x23\x34\x45\x56\x67\x78\x89'
    __OFFLINE_MESSAGE_DATA_ID = b'\x00\xFF\xFF\x00\xFE\xFE\xFE\xFE\xFD\xFD\xFD\xFD\x12\x34\x56\x78'
    __CLIENT_GUID = b'\x00\x00\x00\x00\x00\x00\x00\x00'

    async def get_status(self) -> dict:
        request = self.__ID_UNCONNECTED_PING + self.__TIMESTAMP + self.__OFFLINE_MESSAGE_DATA_ID + self.__CLIENT_GUID
        response = await SocketAsync.send_and_receive(self._host, self._port, self._timeout, request)

        br = BinaryReader(response)
        header = br.read_bytes(1)

        if header != self.__ID_UNCONNECTED_PONG:
            raise InvalidPacketException(
                'Packet header mismatch. Received: {}. Expected: {}.'
                .format(header, self.__ID_UNCONNECTED_PONG)
            )

        br.read_bytes(len(self.__TIMESTAMP) + len(self.__CLIENT_GUID))  # skip timestamp and guid
        magic = br.read_bytes(len(self.__OFFLINE_MESSAGE_DATA_ID))

        if magic != self.__OFFLINE_MESSAGE_DATA_ID:
            raise InvalidPacketException(
                'Magic value mismatch. Received: {}. Expected: {}.'
                .format(magic, self.__OFFLINE_MESSAGE_DATA_ID)
            )

        br.read_short()  # skip remaining packet length

        d = [b';']  # delimiter
        result = {}
        result['edition'] = br.read_string(d)
        result['motd_line_1'] = br.read_string(d)
        result['protocol_version'] = int(br.read_string(d))
        result['version'] = br.read_string(d)
        result['num_players'] = int(br.read_string(d))
        result['max_players'] = int(br.read_string(d))

        # Try to read all data
        try:
            result['server_uid'] = br.read_string(d)
            result['motd_line_2'] = br.read_string(d)
            result['gamemode'] = br.read_string(d)
            result['gamemode_numeric'] = int(br.read_string(d))
            result['port_ipv4'] = int(br.read_string(d))
            result['port_ipv6'] = int(br.read_string(d))
        except Exception:
            pass

        return result


if __name__ == '__main__':
    import asyncio
    import json

    async def main_async():
        raknet = Raknet(host='193.70.94.83', port=19132, timeout=5.0)
        status = await raknet.get_status()
        print(json.dumps(status, indent=None) + '\n')

    asyncio.run(main_async())
